(*****************************************************************************
 *                                                                           *
 *                    :: Semestralni prace VT1A ::                           *
 *                                                                           *
 *  Navod:                                                                   *
 *                                                                           *
 *    o   Prevedeni souboru do WAVu:                                         *
 *           nok_play -t 120 soubor.mel -o output.wav                        *
 *    o   Prehrani souboru na PC Speakeru                                    *
 *           nok_play -t 120 soubor.mel -s                                   *
 *                                                                           *
 *                                                                           *
 *****************************************************************************)

Program Nokia_melody_player;

Uses CRT;

(*********************
 *                   *
 * K o n s t a n t y *
 *                   *
 *********************)

Const 
  WAV_LENGHT_SEEK = $28;
(*
 *
 *  Hlavicka pro M$ RIFF WAVE format - 8 bit, 8000 Hz
 *  http://home.attbi.com/~chris.bagwell/AudioFormats-11.html#ss11.6 
 *
 *  Posledni 4 byty obsahuji (long int) delku  PCM dat, ktera jsou pred
 *  zacatkem nahravani neznama. Offset je 0x28, delka hlavicky je 0x2C 
 *
 *)
  RIFF_WAVE_HEADER : array [1..$2C] Of byte = (
                                               $52, $49, $46, $46, $C2, $B8, $00, $00, $57
                                               , $41, $56, $45, $66, $6D, $74,
                                               $20, $10, $00, $00, $00, $01, $00, $01, $00
                                               , $40, $1F, $00, $00, $40, $1F,
                                               $00, $00, $01, $00, $08, $00, $64, $61, $74
                                               , $61, $99, $99, $99, $99
                                              );
{ Maximalni velikost pole }
  MAX = 100;

  TAB = 9;  { Tabelator ASCII #9 }
  BELL = 7; { Pipnuti   ASCII #7 }

  M_PI = 3.14159265358979323846; { Konstana PI }

  Help : array [1..11] Of string[80] = (
                                        '',
                                        'RingPlay, (c) brucham@fel.cvut.cz',
                                        'Use: nok_play <options> file',
                                        '',
                                        '   Options:',
                                        '',
                                        '     -v be versbose',
                                        '     -t tempo (30 - 300), default = 120',
                                        '     -o Output audio to WAV file (8bit, 8000 Hz)'
                                        ,
 (*	'     -p Output to RAW stdout  (8bit, 8000 Hz)', *)
                                        '     -s play on PC Speaker',
                                        ''
                                       );

 (* ** Hudebni konstanty ** *)

 { e^(1/12*ln(2) - kvocient, ktery odstupnovava radu hudnebnich tonu }
  Q = 1.059463094359295264561825294945;
 { Kmitocet komorniho a }
  A4 = 435.00;
 { Kmitocet prvniho tonu ve stupnici }
  C0 = A4/(25*Q);
 { Tempo }
  D = 2.0;
  DEFAULT_TEMPO = 120;
 { D/A nastaveni }
  SAMPLE_RATE = 8000.0;
  BIT_PER_SAMPLE = 8;
  MAX_VAL = 128; { ( ( power(2, BIT_PER_SAMPLE) - 1 ) / 2 ) }

(*********************
 *                   *
 *  D a t.  t y p y  *
 *                   *
 *********************)

Type 
  Tone = Record
           duration : LongInt;
           note : string[3];
           semi_note : boolean;
           semi_duration : byte;
           octave : LongInt;
  End;

 SArray = array[0..MAX] Of string[6];

(*********************
 *                   *
 *  P r o m e n n e  *
 *                   *
 *********************)

Var 
  input_file    :  string;
  output_file   :  string;
  tempo         :  word;
  pulse_shape   :  byte;
  format, index :  byte;
  verbosity     :  byte;
  melody_array  :  SArray;
  tmp_array     :  SArray;
  a, count      :  integer;
  datafile      :  file;
  file_size     :  longint;
  speaker       :  boolean;
    music         :  Tone;

(*********************
 *                   *
 *    F u n k c e    *
 *                   *
 *********************)

Function StrLen(chunk : String) : byte;
{ Vraci delku znaku, ktera je v Pascalu standardne ulozena,
  na nultem indexu }
 Begin
   StrLen := Ord(chunk[0]);
 End;

Function StrScan(chunk : String; find : char) : byte;
{ Vraci pozici prvniho vyskytu znaku v retezci jako jeho index
  od zacatku }
Begin
  StrScan := Pos(find,chunk);
End;

Function StrClr(chunk : String) : string;
 Begin
   chunk[0] := Chr(0);
   StrClr := chunk;
 End;

Function Atoi( cNum: String ): LongInt;
{ Prevede retezec na cislo, obdoba atoi z C }
{ Retezec nesmi obsahovat dalsi znaky napr. mezeru,
  jinak to funguje spatne }

 Var 
   c : integer;
   i : LongInt;
 Begin
             Val(cNum, i, c );
     Atoi := i;
 End;


Function Exists (name : String) : boolean;
{ Zjisteni existence souboru }

 Var fm : Byte;
   f  : File;
   b  : Boolean;
 Begin
             {$I-}
   fm := FileMode;
   FileMode := 0;
   assign (f, name);
   reset(f);
   b := IOResult = 0;
                {$I+}
   If b Then close(f);
   Exists := b;
   FileMode := fm;
 End;


Function Param_file_name : string;
{ Precteni jmena souboru ze vstupnich parametru }

 Var i : LongInt;
   param : string;
 Begin
   For i := 1 To ParamCount Do
     Begin
       param := ParamStr(i);
       If (param[1] <> '-') Then
         If (Exists(param)) Then
           Begin
             Param_file_name := param;
             break;
           End;
     End;
 End;


Procedure Param_Parser;
{ Nastaveni promenych na zaklade vstupni radky }

 Var 
   argc  : string;
   i : integer;
 Begin
   output_file := '';
   For i := 1 To ParamCount Do
     Begin
       argc := ParamStr(i);
       If (argc[1] = '-') Then
         Begin
           Case argc[2] Of
     { Ukecanost vystupu }
             'v','V' : verbosity := verbosity + 1;
     { Tempo }
             't','T' : tempo := Atoi(ParamStr(i+1));
     { Nokia }
             'n','N' : format := 0;
     { Tvar amplitudy }
             'a','A' : pulse_shape := Atoi(ParamStr(i+1));
     { Vystupni soubor }
             'o','O' : output_file := ParamStr(i+1);
             'p','P' : output_file := 'stdout'; { na obrazovku }
        { Vystup na buzzer }
             's','S' : speaker := TRUE;
             Else
               Begin
                 writeln(Chr(BELL), 'Unknown option: ', argc[2],' !');
               End;
           End;
         End;
     End;

  { prednastavene hodnoty }
   If ( tempo = 0 ) Then
     tempo := DEFAULT_TEMPO;
   If ( StrLen(output_file) = 0 ) Then
     output_file := '';
 End;

Function Explode(sentence : String; separator : char): integer;
{ Rozdeli retezec podle urciteho separatoru na podretezce, vice separatoru
  vedle sebe je povazovano za jeden }

 Var 
   j : integer;
   sep: boolean;
   size,k,i : byte;

 Begin
   size := ord(sentence[0]);
   j := -1;  {pocet slov, resp. ukazatel na zapisovane slovo}
   k := 0;   {ukazatel na aktualni pozici ve slove}
   sep := TRUE;   { pred slovem je logicky separator }
   For i := 1 To size Do
     Begin
       If (sentence[i] <> separator) Then { jsme ve slove }
         Begin
           If (sep) Then { pokud predchozi byl separator, pak je nove slovo }
             Begin
               sep := FALSE;
               j := j + 1;
             End;
    {kazdopadne musime zaznamenat znak}
           k := k + 1;
           tmp_array[j][k] := sentence[i];
         End
       Else
         Begin
           If (Not sep) Then {pokud je prvni separator, ukoncime predchozi slovo}
             Begin
               sep := TRUE;
               tmp_array[j][0] := chr(k);
               k := 0;
             End;
         End;
     End;
   If (Not sep) Then {poku posledni nebyl separator, ukoncime predchozi slovo}
     tmp_array[j][0] := chr(k);

   Explode := j+1;
 End;

Procedure Read_file(input_file : String);
{ Nacte jednotlive polozky ze souboru oddelene ' ' a nastavi hodnotu globalni
  promenne count na jejich pocet }

 Var 
   buffer        :  string;
   music_data    :  text;
   u, c          :  byte;
   words  : longint;
 Begin
   Assign(music_data, input_file);
   Reset(music_data);

   words := 0;

   While not Eof(music_data) Do
   Begin
     readln(music_data, buffer);
{			writeln(buffer); }
     count := Explode(buffer, ' ');

     words := words + count;
{			writeln('Count: ',count,' Words: ',words);}

     If ( count > 0 ) Then
       Begin
         c := 0;
         For u := index To (index + count) Do
           Begin
{					writeln(u,'	|',tmp_array[c],'|');}
             melody_array[u] := tmp_array[c];
             c := c + 1;
           End;
         index := u;
       End;
   End;

   Close(music_data);
   count := words;
 End;


Procedure Print_help;

 Var 
   a : integer;
 Begin
   For a := 1 To ( Sizeof(Help) Div ( Sizeof(char) * 80 ) ) Do
       writeln(Help[a]);
  End;

Function IsAlpha(letter : char) : boolean;
{ Test zda znak je znak a-z nebo A-Z }
 Begin
   letter := upcase(letter);
   If (letter >= 'A') And (letter <= 'Z') Then
     IsAlpha := TRUE
   Else
     IsAlpha := FALSE;
 End;

Function IsDigit(letter : char) : boolean;
{ Test zda znak je ci neni cislem }
 Begin
   If (letter >= '0') And (letter <= '9') Then
     IsDigit := TRUE
   Else
     IsDigit := FALSE;
 End;

(* ************************************************************ *)

Procedure parse_nokia_composer(xtone : String);
{ Zpracuje data obsahujici popis melodie ve formatu Nokia Composer }

 Var 
   i, j, str_lenght : byte;
   buffer : string[3];
 Begin
   str_lenght := StrLen(xtone);
   buffer := '';

  { Udaj o oktave se nachazi na podlednim miste }
   music.octave := atoi(xtone[str_lenght]) + 3;

  { Udaj o delce tonu 1, 2, 4, 8, 16, 32 }
   For i := 1 To str_lenght Do
     Begin
       If (IsDigit(xtone[i])) Then
         Begin
           buffer[i] := xtone[i];
         End
       Else
         break;
     End;
   buffer[0] := Chr(i - 1);
   music.duration := atoi(buffer); { OK }

  { Udaj o tonu a - g, #xtone - pulova, - pausa }
  {pokud ma byt spravne i:=1 (0 je delka retezce), pak je j zbytecny}
   j := 1;
   For i := i To str_lenght Do
     Begin
       If ( IsAlpha(xtone[i]) Or (xtone[i] = '#') Or (xtone[i] = '-')) Then
         Begin
           buffer[j] := xtone[i];
           j := j + 1;
         End
       Else
         break;
     End;
   buffer[0] := Chr(j - 1);
   music.note := buffer; { OK }
          End;

Function Power(base : real; exponent : real) : Real;
{ Funkce Power neni definovana v Borland Pascalu }
Begin
     Power := exp( exponent * ln(base) );
End;

Function tone2frequency(music_data : Tone) : Real;
{ Vrati frekvenci urciteho tonu }

 Var 
   oct : integer;
   freq : Real;
   semitone : Real;
   xtone : char;
 Begin
   oct := music_data.octave;

   If (music_data.note[1] = '#') Then
     Begin
       xtone := music_data.note[2];
       semitone := Q;
     End
   Else
     Begin
       xtone := music_data.note[1];
       semitone := 1.0;
     End;

   Case xtone Of
     'c':
          freq := C0 * Power(2.0, oct) * semitone;
     'd':
          freq := C0 * Power(Q, 2.0) * Power(2.0, oct) * semitone;
     'e':
          freq := C0 * Power(Q, 4.0) * Power(2.0, oct) * semitone;
     'f':
          freq := C0 * Power(Q, 5.0) * Power(2.0, oct) * semitone;
     'g':
          freq := C0 * Power(Q, 7.0) * Power(2.0, oct) * semitone;
     'a':
          freq := C0 * Power(Q, 9.0) * Power(2.0, oct) * semitone;
     'b':
          freq := C0 * Power(Q, 11.0) * Power(2.0, oct);
     Else
       freq := 2;
   End;

   tone2frequency :=  freq;
 End;

Procedure Speaker_Play(freq : real; tdelay : real);
 Begin
  { POZOR: * rutina sound neni implementovana v Linuxu!! }
  {        * ve Windows prehraje systemove pipnuti a frekvenci ignoruje }
   sound(round(freq));
   delay(round(tdelay));
   nosound;
 End;

Procedure Play(freq : real; tdelay : real);

 Var 
   i : real;
   timer : LongInt;
                data : byte;
 Begin
   i := 0;
   For timer := 0 To round( (tdelay/1000)/((1/SAMPLE_RATE)) ) Do
     Begin
       Repeat
  {	write(datafile, Chr(Byte(round(sin(i)*(MAX_VAL-1)+MAX_VAL)))); }
   { zapis PCM dat s obdelnikovou modulaci }
         If (sin(i) > 0) Then
           Begin
             If (output_file = 'stdout') Then
               Begin
                 write(Chr(2 * MAX_VAL - 1));
               End
             Else
               Begin
                    data := (2 * MAX_VAL) - 1;
                 Blockwrite(datafile, data, 1);
               End;
           End
         Else
           If (output_file = 'stdout') Then
             Begin
               write(Chr(0));
             End
         Else
           Begin
                    data := 0;
             Blockwrite(datafile, data, 1);
           End;
         i := i + (2*M_PI)/( SAMPLE_RATE/freq );
       Until ( i >= 2 * M_PI );
     End;
End;

Procedure mpause(tdelay : real);

 Var 
   timer : LongInt;
                data : byte;
 Begin
   For timer := 0 To round( (tdelay/1000)/((1/SAMPLE_RATE)) ) Do
     Begin
                        data := MAX_VAL;
       Blockwrite(datafile, data, 1);
     End;
 End;

Procedure Spause(tdelay : real);
 Begin
   delay(round(tdelay));
 End;

Procedure play_tone(music_data : Tone; tempo : real);
{ Prehraje ton rychlosti podle tempa }

 Var 
   freq, tdelay : real;
 Begin
   tempo := tempo / 60;
   tdelay := round(( ( D / ( music_data.duration * tempo ) )*1000));
   If ( (music_data.note[1] = '-') Or (music_data.note[1] = 'p') ) Then
     Begin
       If ( speaker = TRUE ) Then
         Begin
           Spause(tdelay);
         End
       Else
         mpause(tdelay);
     End
     Else
       Begin
         freq := tone2frequency(music_data);
         If ( speaker = TRUE ) Then
           Begin
             Speaker_Play(freq, tdelay);
           End
         Else
           play(freq, tdelay);
       End;
 End;

Procedure Print_info;
Begin
  write(melody_array[a],' :: ');
  write(' n: ', music.note);
  write(' d: ','1/',music.duration);
  write(' o: ', music.octave);
  write(' f = ', tone2frequency(music): 5: 2 , ' Hz');
  writeln;
End;


(*********************
 *                   *
 *   P r o g r a m   *
 *                   *
 *********************)

Begin
  tempo := 0;
 { Zpracovani argumentu prikazove radky }
  If (ParamCount < 1) Then
    Begin
       Print_help;
      exit;
    End;

        {$I-}
  input_file := Param_file_name;
        {$I+}

  If (input_file = '') Then
    Begin
      writeln(Chr(BELL),'Cannot open input file!');
      exit;
    End;
  Param_Parser;

 { Nacteni vstupniho souboru }
  Read_file(input_file);

  If (verbosity > 0) Then
    Begin
      writeln('Input file: ', input_file);
      writeln('Output file: ', output_file);
      Writeln('Verbosity: ', verbosity);
      Writeln('Format: ', format);
      Writeln('Tempo: ', tempo);
      Writeln('Number of notes: ', count);
      Writeln;
    End;

 (* ****** Prehrani na speakeru ****** *)

  If (speaker = TRUE) Then
    Begin
      writeln('Listen the PC Speaker.');
      For a := 0 To (count - 1)  Do
        Begin
          parse_nokia_composer(melody_array[a]);

                        If (verbosity > 0) Then
                          Print_info;

                        If ( (music.note = '') Or
                             (music.duration = 0) ) Then
                             Begin
                                  writeln(Chr(BELL),'Invalid data at item no. ',a);
                                  exit
                             End
                        Else
                          play_tone(music, tempo);
                End;
    End;

 (* ********* Zapis do WAVu ********** *)

  If ( (strlen(output_file) > 0)  And (output_file <> 'stdout') ) Then
    Begin
      writeln('Writing...');
      assign(datafile, output_file);
      rewrite(datafile, sizeof(byte));

  { zapsani hlavicky }
      For a := 1 To sizeof(RIFF_WAVE_HEADER) Do
        Begin
          blockwrite(datafile, RIFF_WAVE_HEADER[a], 1);
        End;
  { zapsani PCM dat }
      For a := 0 To (count - 1)  Do
        Begin
            parse_nokia_composer(melody_array[a]);
          play_tone(music, tempo);

            If (verbosity > 0) Then
              Print_info;
        End;
  { zapsani velikosti zapsanych dat do hlavicky }
      file_size :=  (Filesize(datafile) * SizeOf(byte)) - sizeof(RIFF_WAVE_HEADER);
      Seek(datafile, WAV_LENGHT_SEEK);
  { zapsani LongInt po bytech na Little Endian architekture }
      pulse_shape := Byte( file_size And $000000FF);
      blockwrite(datafile, pulse_shape, 1);
      pulse_shape := Byte((file_size And $0000FF00) shr 8);
      blockwrite(datafile, pulse_shape, 1);
      pulse_shape := Byte((file_size And $00FF0000) shr 16);
      blockwrite(datafile, pulse_shape, 1);
      pulse_shape := Byte((file_size And $FF000000) shr 32);
      blockwrite(datafile, pulse_shape, 1);
      close(datafile);

      writeln(file_size + sizeof(RIFF_WAVE_HEADER), ' bytes written into ', output_file);
    End;
  writeln;
End.
ln('Listen the PC Speaker.');
            For a := 0 To (count - 1)  Do
                Begin
                    parse_nokia_composer(melody_array[a]);

                        If (verbosity > 0) Then
                            Print_info;

                        If ( (music.note = '') Or
                             (music.duration = 0) ) Then
                             Begin
                                  writeln(Chr(BELL),'Invalid data at item no. ',a);
                                  exit
                             End
                        Else
                            play_tone(music, tempo);
                End;
        End;

 (* ********* Zapis do WAVu ********** *)

    If ( (strlen(output_file) > 0)  And (output_file <> 'stdout') ) Then
        Begin
            writeln('Writing...');
            assign(datafile, output_file);
            rewrite(datafile, sizeof(byte));

  { zapsani hlavicky }
            For a := 1 To sizeof(RIFF_WAVE_HEADER) Do
                Begin
                    blockwrite(datafile, RIFF_WAVE_HEADER[a], 1);
                End;
  { zapsani PCM dat }
            For a := 0 To (count - 1)  Do
                Begin
                    parse_nokia_composer(melody_array[a]);
                    play_tone(music, tempo);

                    If (verbosity > 0) Then
                        Print_info;
                End;
  { zapsani velikosti zapsanych dat do hlavicky }
            file_size :=  (Filesize(datafile) * SizeOf(byte)) - sizeof(RIFF_WAVE_HEADER);
            Seek(datafile, WAV_LENGHT_SEEK);
  { zapsani LongInt po bytech na Little Endian architekture }
            pulse_shape := Byte( file_size And $000000FF);
            blockwrite(datafile, pulse_shape, 1);
            pulse_shape := Byte((file_size And $0000FF00) shr 8);
            blockwrite(datafile, pulse_shape, 1);
            pulse_shape := Byte((file_size And $00FF0000) shr 16);
            blockwrite(datafile, pulse_shape, 1);
            pulse_shape := Byte((file_size And $FF000000) shr 32);
            blockwrite(datafile, pulse_shape, 1);
            close(datafile);

            writeln(file_size + sizeof(RIFF_WAVE_HEADER), ' bytes written into ',
            output_file);
        End;
    writeln;
End.
